import pygame

"""
	Classe base para botões.
"""
class Botao():
	def __init__(self, posicao, ativado):
		self.posicao = posicao
		self.ativado = ativado
		self.GerarBotao()

	def draw(self, tela):
		pass

	# Verifica se o botão foi apertado
	def Verificar(self, apertado):
		if self.rect.collidepoint(pygame.mouse.get_pos()) and self.ativado:
			if pygame.mouse.get_pressed()[0] and not apertado:
				return True
		return False


"""
	Botão com texto.
"""
class BotaoTexto(Botao):
	def __init__(self, posicao, texto, fonte, cor_texto, cor_fundo, cor_borda, cor_bloqueado, ativado):
		self.texto = texto
		self.fonte =  fonte
		self.corTexto = cor_texto
		self.corFundo = cor_fundo
		self.corBorda = cor_borda
		self.corBloqueado = cor_bloqueado
		Botao.__init__(self, posicao, ativado)

	def draw(self, tela):
		cor_rect = self.corFundo if self.ativado else self.corBloqueado
		pygame.draw.rect(tela, cor_rect, self.rect)
		pygame.draw.rect(tela, self.corBorda, self.rect, width=2)
		tela.blit(self.texto_desenho, (self.rect.left + 10, self.rect.top + 10))

	# Gera o texto e a caixa do botão
	def GerarBotao(self):
		self.texto_desenho = self.fonte.render(self.texto, True, self.corTexto)
		texto_alt = self.texto_desenho.get_height()
		texto_larg = self.texto_desenho.get_width()
		self.rect = pygame.Rect(0, 0, texto_larg + 20, texto_alt + 20)
		self.rect.center = (self.posicao[0], self.posicao[1])

    ### Setters
	# atualiza o texto do botão
	def SetTexto(self, texto):
		self.texto = texto
		self.GerarBotao()
    
	# atualiza a posição do botão
	def SetPos(self, posicao):
		self.posicao = posicao
		self.GerarBotao()
    
	# atualiza a fonte do botão
	def SetFonte(self, fonte):
		self.fonte = fonte
		self.GerarBotao()
        
	# atualiza a cor do texto do botão
	def SetCorTexto(self, cor_texto):
		self.corTexto = cor_texto
		self.GerarBotao()

	# atualiza a cor do botão
	def SetCorBotao(self, cor_fundo, cor_borda, cor_bloqueado):
		self.corFundo = cor_fundo
		self.corBorda = cor_borda
		self.corBloqueado = cor_bloqueado

	# ativa ou desativa o botão
	def setAtivado(self, ativado):
		self.ativado = ativado
		

"""
	Classe do botão com imagem.
"""
class BotaoImagem(Botao):
	def __init__(self, posicao, imagem, escala, cor_borda, ativado):
		alt = imagem.get_height()
		larg = imagem.get_width()
		self.imagem = pygame.transform.scale(imagem, (int(larg * escala), int(alt * escala)))
		self.corBorda = cor_borda
		Botao.__init__(self, posicao, ativado)

	def draw(self, tela):
		tela.blit(self.imagem, self.rect)
		pygame.draw.rect(tela, self.corBorda, self.rect, width=2)
	
	# gera o botão baseado na imagem
	def GerarBotao(self):
		self.rect = self.imagem.get_rect()
		self.rect.center = self.posicao

	### Setters
	# atualiza a imagem
	def SetImagem(self, imagem, escala):
		alt = imagem.get_height()
		larg = imagem.get_width()
		self.imagem = pygame.transform.scale(imagem, (int(larg * escala), int(alt * escala)))
		self.GerarBotao()

	# atualiza a escala da imagem
	def SetEscala(self, escala):
		alt = self.imagem.get_height()
		larg = self.imagem.get_width()
		self.imagem = pygame.transform.scale(self.imagem, (int(larg * escala), int(alt * escala)))
		self.GerarBotao()

	# atualiza a posição do botão
	def SetPos(self, posicao):
		self.posicao = posicao
		self.GerarBotao()

	# atualiza a cor da borda do botão
	def SetCor(self, cor_borda):
		self.corBorda = cor_borda



'''
tipo:
	0 = float
	1 = int
	2 = string
'''
"""
	Classe da caixa de texto
"""
class CaixaTexto():
	def __init__(self, posicao, tamanho, nome, fonte_nome, fonte_texto, tipo, valor_base, tamanho_maximo, cor_nome, cor_texto, cor_fundo, cor_borda, cor_selecionado):
		self.posicao = posicao
		self.tamanho = tamanho
		self.nome = nome
		self.fonteTexto = fonte_texto
		self.fonteNome = fonte_nome
		self.tipo = tipo
		self.valorPadrao = valor_base
		self.valorBase = valor_base
		self.valorAtual = valor_base
		self.tamanhoMaximo = tamanho_maximo
		self.corNome = cor_nome
		self.corTexto = cor_texto
		self.corFundo = cor_fundo
		self.corBorda = cor_borda
		self.corSelecionado = cor_selecionado
		self.digitando = False
		self.GerarTexto()
		self.GerarCaixa()

	def draw(self, tela):
		cor_rect = self.corSelecionado if self.digitando else self.corFundo
		pygame.draw.rect(tela, cor_rect, self.rect)
		pygame.draw.rect(tela, self.corBorda, self.rect, width=2)
		tela.blit(self.texto_nome, (self.posicao[0] - (self.texto_nome.get_width()/2), 
									self.posicao[1] - (self.texto_nome.get_height() + (self.rect.height/2) + 5)))
		texto_tam = (self.texto_desenho.get_width(), self.texto_desenho.get_height())
		pos_texto = ((self.rect.width - texto_tam[0])/2 + self.rect.left, (self.rect.height - texto_tam[1])/2 + self.rect.top)
		tela.blit(self.texto_desenho, pos_texto)

	# Verifica se a caixa foi selecionada e atualiza o texto baseado na entrada do teclado
	def Verificar(self, apertado, teclas, ativo):
		if self.rect.collidepoint(pygame.mouse.get_pos()) and not self.digitando and not ativo:
			if pygame.mouse.get_pressed()[0] and not apertado:
				self.digitando = True
				self.valorAtual = ''	
				return 1	
		elif self.digitando:
			for tecla in teclas:
				if tecla.key == pygame.K_BACKSPACE:
					if len(self.valorAtual) > 0:
						self.valorAtual = self.valorAtual[:-1]
				elif tecla.key == pygame.K_RETURN:
					self.digitando = False
					if len(self.valorAtual) < 1:
						self.valorAtual = self.valorBase
					else:
						match self.tipo:
							case 0:
								try:
									float(self.valorAtual)
								except ValueError:
									self.valorAtual = self.valorBase
							case 1:
									try:
										int(self.valorAtual)
									except ValueError:
										self.valorAtual = self.valorBase
					return -1
				elif len(self.valorAtual) < self.tamanhoMaximo:
					self.valorAtual += tecla.unicode
		self.GerarTexto()
		return 0

	# gera o texto que será desenhado na tela
	def GerarTexto(self):
		self.texto_nome = self.fonteNome.render(self.nome, True, self.corNome)
		self.texto_desenho = self.fonteTexto.render(self.valorAtual, True, self.corTexto)

	# gera caixa de texto
	def GerarCaixa(self):
		self.rect = pygame.Rect(0, 0, self.tamanho, self.texto_desenho.get_height() + 10)
		self.rect.center = (self.posicao[0], self.posicao[1])

	# volta o texto da caixa para um texto base
	def reset(self):
		self.valorAtual = self.valorBase

	##Setter
	# atualiza a posição da caixa
	def SetPos(self, posicao):
		self.posicao = posicao
		self.GerarCaixa()

	# atualiza o tamanho da caixa
	def SetTam(self, tamanho):
		self.tamanho = tamanho
		self.GerarCaixa()

	# atualiza o tipo de entrada da caixa
	def SetTipo(self, tipo):
		self.tipo = tipo
	
	# atualiza oo texto base da caixa
	def SetValor(self, valor_base):
		self.valorBase = valor_base
		self.reset()

	# atualiza a cor do texto
	def SetCorTexto(self, cor_texto):
		self.corTexto = cor_texto
		self.GerarTexto()
	
	# atualiza a cor da caixa
	def SetCor(self, cor_fundo, cor_borda, cor_selecionado):
		self.corFundo = cor_fundo
		self.corBorda = cor_borda
		self.corSelecionado = cor_selecionado

	# atualiza o rótulo da caixa
	def SetNome(self, nome, fonte, cor):
		self.nome = nome
		self.fonteNome = fonte
		self.corNome = cor
		self.GerarTexto()

	##Getter
	# retorna o texto da caixa
	def getValor(self):
		return self.valorAtual
	
class BotaoSelecao():
	def __init__(self, posicao, texto, fonte, cor_texto, cor_fundo, cor_borda, cor_selecionado, cor_bloqueado, ativado):
		self.texto = texto
		self.fonte =  fonte
		self.corTexto = cor_texto
		self.corFundo = cor_fundo
		self.corBorda = cor_borda
		self.corSelecionado = cor_selecionado
		self.corBloqueado = cor_bloqueado
		self.ativo = False
		self.posicao = posicao
		self.ativado = ativado
		self.GerarBotao()
	
	def draw(self, tela):
		cor_rect = self.corSelecionado if self.ativo else self.corFundo
		pygame.draw.rect(tela, cor_rect, self.rect)
		pygame.draw.rect(tela, self.corBorda, self.rect, width=2)
		tela.blit(self.textoDesenho, (self.posicao[0] - (self.textoDesenho.get_width()/2), 
									  self.posicao[1] - (self.textoDesenho.get_height() + (self.rect.height/2) + 5)))
	
	# gera o botão baseado na imagem
	def GerarBotao(self):
		self.textoDesenho = self.fonte.render(self.texto, True, self.corTexto)
		self.rect = pygame.Rect(0, 0, 30, 30)
		self.rect.center = (self.posicao[0], self.posicao[1])
	
	def Verificar(self, apertado):
		if self.rect.collidepoint(pygame.mouse.get_pos()) and self.ativado:
			if pygame.mouse.get_pressed()[0] and not apertado:
				self.ativo = False if self.ativo else True
	
	def SetValor(self, valor):
		self.ativo = valor

	def GetValor(self):
		return self.ativo


# desenha o texto fornecido na tela
def DesenhaTexto(tela, texto, fonte, cor, pos):
	texto_desenho = fonte.render(texto, True, cor)
	alt = texto_desenho.get_height()
	larg = texto_desenho.get_width()
	tela.blit(texto_desenho, (pos[0]-(larg/2),pos[1]-(alt/2)))

