import pygame
import math
from utilidade import Linha
#Tamanho das linhas de visão
TAMANHO_LINHA = 300


"""
    A classe do carro usado no simulador.
"""
class Carro:
    def __init__(self, pos, angulo, max_vel, vel_rotacao, acc, imagem=None):
        self.imagem = imagem
        self.rect = self.imagem.get_rect()
        self.rect.move_ip(pos[0], pos[1])
        self.vel = 0
        self.max_vel = max_vel
        self.vel_rotacao = vel_rotacao
        self.pos = [pos[0], pos[1]]
        self.angulo = angulo
        self.aceleracao = acc
        self.vector = [self.rect.topleft, self.rect.topright, self.rect.bottomright, self.rect.bottomleft]
        aux = [(self.vector[0][0] + self.vector[1][0])/2, (self.vector[0][1] + self.vector[1][1])/2]
        aux2 = [(self.vector[2][0] + self.vector[3][0])/2, (self.vector[2][1] + self.vector[3][1])/2]
        self.linhas = [Linha(aux, TAMANHO_LINHA, 0, self.angulo), Linha(aux, TAMANHO_LINHA, 30, self.angulo), Linha(aux, TAMANHO_LINHA, 60, self.angulo), 
                      Linha(aux, TAMANHO_LINHA, 90, self.angulo), Linha(aux, TAMANHO_LINHA, 120, self.angulo), Linha(aux, TAMANHO_LINHA, 150, self.angulo), 
                      Linha(aux, TAMANHO_LINHA, 180, self.angulo), Linha(aux, TAMANHO_LINHA, 225, self.angulo), Linha(aux, TAMANHO_LINHA, 270, self.angulo), 
                      Linha(aux, TAMANHO_LINHA, 315, self.angulo), ]
        self.linha_atras = 3
    
    def draw(self, win):
        self.AtualizaVector()
        imagem_rotacionada = pygame.transform.rotate(self.imagem, self.angulo)
        novo_rect = imagem_rotacionada.get_rect(center=self.imagem.get_rect(topleft=(self.pos)).center)
        win.blit(imagem_rotacionada, novo_rect.topleft)
        for linha in self.linhas:
            linha.draw(win)

    # atualiza os pontos da caixa de colisão do carro
    def AtualizaVector(self):
        pivo = pygame.math.Vector2(self.rect.center)
        pts = [self.rect.topleft, self.rect.topright, self.rect.bottomright, self.rect.bottomleft]
        pts = [(pygame.math.Vector2(p) - pivo).rotate(-self.angulo) + pivo for p in pts]
        
        self.vector = pts
        for i in range(0, len(self.linhas)-self.linha_atras):
            self.linhas[i].MovePonto([(self.vector[0][0] + self.vector[1][0])/2, 
                                      (self.vector[0][1] + self.vector[1][1])/2])
        for i in range(len(self.linhas)-self.linha_atras, len(self.linhas)):
            self.linhas[i].MovePonto([(self.vector[2][0] + self.vector[3][0])/2, 
                                      (self.vector[2][1] + self.vector[3][1])/2])

    # recebe e realiza a ação escolhida pelo agente
    def Entrada(self, acao):

        if acao[0] == 1:
            self.Acelerar()

        elif acao[1] == 1:
            self.Rotacionar(direita=True)

        elif acao[2] == 1:
            self.Frear()

        elif acao[3] == 1:
            self.Rotacionar(esquerda=True)
            

    # gira o carro para a esquerda ou direita
    def Rotacionar(self, esquerda=False, direita=False):
        if esquerda:
            self.angulo += self.vel_rotacao
            for linha in self.linhas:
                linha.SomaAngulo(self.vel_rotacao)
        elif direita:
            self.angulo -= self.vel_rotacao
            for linha in self.linhas:
                linha.SomaAngulo(-self.vel_rotacao)
        self.vel = max(self.vel - self.aceleracao/2, self.max_vel/4)
        self.Mover()
        self.AtualizaVector()

    # aumenta a velocidade do carro caso não esteja chegado na velocidade máxima
    def Acelerar(self):
        self.vel = min(self.vel + self.aceleracao, self.max_vel)
        self.Mover()

    # desacelera o carro até chegar a velocidade mínima
    def Frear(self):
        self.vel = max(self.vel - self.aceleracao, -self.max_vel/2)
        self.Mover()

    # move o carro baseado em sua velocidade e ángulo
    def Mover(self):
        radianos = math.radians(self.angulo)
        vertical = math.cos(radianos) * self.vel
        horizontal = math.sin(radianos) * self.vel
        self.pos[1] -= vertical
        self.pos[0] -= horizontal
        self.rect.topleft = (self.pos)

        self.AtualizaVector()

    # restaura a posição e ángulo do carro
    def reset(self, pos, angulo):
        self.pos = [pos[0], pos[1]]
        self.rect.topleft = (self.pos)
        self.angulo = angulo
        self.vel = 0
        self.vector = [self.rect.topleft, self.rect.topright, self.rect.bottomright, self.rect.bottomleft]
        aux = [(self.vector[0][0] + self.vector[1][0])/2, (self.vector[0][1] + self.vector[1][1])/2]
        aux2 = [(self.vector[2][0] + self.vector[3][0])/2, (self.vector[2][1] + self.vector[3][1])/2]
        for i in range(0, len(self.linhas)-self.linha_atras):
            self.linhas[i].reset(aux, self.angulo)
        for i in range(len(self.linhas)-self.linha_atras, len(self.linhas)):
            self.linhas[i].reset(aux2, self.angulo)