import pygame
import menu
import glob
import json
import os
import simulador

"""
	Controlador Principal do Projeto
"""


#iniciando Pygame
pygame.init()
pygame.font.init()
FPS = 60
LARGURA, ALTURA = 1280, 720
TELA = pygame.display.set_mode((LARGURA, ALTURA))
pygame.display.set_caption("Simulador de corrida")

CAMINHO_MODELO = "./modelos/"

"""retorna a quantidade de pistas"""
def VerificarPistas()->int:

	f = open("Pistas.json")
	pistas = json.load(f)
	f.close()
	return len(pistas)

"""retorna uma lista com todos os modelos já criados"""
def VerificaModelos(quantidade_pistas)->list:
	modelos = []
	for i in range(quantidade_pistas):
		modelos.append([])
	for pista in glob.glob(f"{CAMINHO_MODELO}/*"):
		if os.path.isdir(pista):
			nome = pista[pista.find("pista")+5:]
			pista_num = int(nome)
			for modelo in glob.glob(f"{pista}/*"):
				if os.path.isdir(modelo):
					modelos[pista_num].append([os.path.basename(modelo), modelo])

	return modelos

"""Guarda qual o ultimo modelo usado em determinada pista"""
def GravaUltModelo(caminho, modelo):
	arquivo = open(caminho, "w")
	arquivo.write(f"{modelo}")
	arquivo.close()


"""controla o funcionamento do projeto"""
def ControladorPrincipal():
	janela = 0
	rodando = True
	quantidade_pistas = VerificarPistas()
	modelos = VerificaModelos(quantidade_pistas)
	objeto_menu = menu.Menu(LARGURA, ALTURA, quantidade_pistas, modelos, CAMINHO_MODELO)
	objeto_simulador = simulador.Simulador(LARGURA, ALTURA)
	# laço principal
	while rodando:
		# caso esteja no menu
		if janela == 0:
			retorno, num_pista, nome_modelo = objeto_menu.update(TELA)
			match retorno:
				case -1:
					rodando = False
				case 1:
					caminho_arquivo = f"{CAMINHO_MODELO}pista{num_pista}/Ultmodelo"
					GravaUltModelo(caminho_arquivo, nome_modelo)
					objeto_simulador.carregarPista(num_pista, nome_modelo, CAMINHO_MODELO, 60)
					janela = 1
		# caso esteja no simulador
		elif janela == 1:
			retorno = objeto_simulador.PassoSimulador(TELA)
			match retorno:
				case -2:
					rodando = False
				case -1:
					janela = 0
	pygame.quit()



if __name__ == '__main__':
    ControladorPrincipal()

