import pygame
import time
import math
import glob
import json
from carro import Carro
import random
import utilidade

import os
import shutil

from modelo import Agente
import botao
import grafico

pygame.font.init()
FONTE = pygame.font.Font('arial.ttf', 25)
FONTE1 = pygame.font.Font('arial.ttf', 50)

COR_FONTE1 = pygame.Color('ivory')
COR_FONTE2 = pygame.Color('antiquewhite')
COR_FUNDO_BOTAO = pygame.Color('darkslategrey')
COR_BORDA_BOTAO = pygame.Color('grey10')
COR_BOTAO_BLOQUEADO = pygame.Color('grey25')

'''
    Classe responsável pelo simulador de corrida e comunicação com o agente.
'''
class Simulador:
    def __init__(self, largura, altura):
        self.clock = pygame.time.Clock()
        self.LARGURA = largura
        self.ALTURA = altura

        self.carro_img = pygame.image.load("carro.png").convert_alpha()
        self.carro = Carro((0,0), 0, 4, 4, 1, self.carro_img)
        self.pontos_col = []
        self.dist_controle = None
        self.dist_controle_maxima = None
        self.pontuacao = 0
        self.contagem_regressiva = 0
        self.pontuacao_total = 0
        self.pontuacao_maxima = 0
        self.terminado = 0
        self.linha = [[0, 0], [0, 0]]
        self.dir_carro = 0
        self.botoes = {}
        self.visualizar = False
        self.apertado = False
        self.mostrar_dados = True
        self.botoes['ini_vis'] = botao.BotaoTexto((305, self.ALTURA-50), "VISUALIZAR PONTUAÇÃO", FONTE, COR_FONTE1, COR_FUNDO_BOTAO, COR_BORDA_BOTAO, COR_BOTAO_BLOQUEADO, True)
        self.botoes['stp_vis'] = botao.BotaoTexto((300, self.ALTURA-50), "PARAR DE VISUALIZAR", FONTE, COR_FONTE1, COR_FUNDO_BOTAO, COR_BORDA_BOTAO, COR_BOTAO_BLOQUEADO, True)
        self.botoes['print'] = botao.BotaoTexto((605, self.ALTURA-50), "GRAVAR GRAFICO", FONTE, COR_FONTE1, COR_FUNDO_BOTAO, COR_BORDA_BOTAO, COR_BOTAO_BLOQUEADO, True)
        self.botoes['esconde'] = botao.BotaoTexto((620, self.ALTURA-50), "ESCONDER DADOS", FONTE, COR_FONTE1, COR_FUNDO_BOTAO, COR_BORDA_BOTAO, COR_BOTAO_BLOQUEADO, True)
        self.botoes['mostra'] = botao.BotaoTexto((620, self.ALTURA-50), "MOSTRAR DADOS", FONTE, COR_FONTE1, COR_FUNDO_BOTAO, COR_BORDA_BOTAO, COR_BOTAO_BLOQUEADO, True)
        self.botoes['trava_qps'] = botao.BotaoTexto((860, self.ALTURA-50), "TRAVAR QPS", FONTE, COR_FONTE1, COR_FUNDO_BOTAO, COR_BORDA_BOTAO, COR_BOTAO_BLOQUEADO, True)
        self.botoes['destrava_qps'] = botao.BotaoTexto((880, self.ALTURA-50), "DESTRAVAR QPS", FONTE, COR_FONTE1, COR_FUNDO_BOTAO, COR_BORDA_BOTAO, COR_BOTAO_BLOQUEADO, True)
        self.botoes['mudar_grafico'] = botao.BotaoTexto((880, self.ALTURA-50), "MUDAR GRÁFICO", FONTE, COR_FONTE1, COR_FUNDO_BOTAO, COR_BORDA_BOTAO, COR_BOTAO_BLOQUEADO, True)
        self.botoes['voltar'] = botao.BotaoTexto((75, self.ALTURA-50), "SAIR", FONTE, COR_FONTE1, COR_FUNDO_BOTAO, COR_BORDA_BOTAO, COR_BOTAO_BLOQUEADO, True)

        self.grafico = grafico.Grafico(self.LARGURA, self.ALTURA)
        self.mostrar_grafico = False
        self.grafico_modificado = False
        self.grafico_sur = pygame.Surface((self.LARGURA,self.ALTURA))
        self.grafico_tipo = 0
        self.travar_qps = False
        
    # Carrega as informações da pista selecionada
    def carregarPista(self, num_pista, modelo, caminho_modelos, qps):
        f = open("Pistas.json")
        self.pistas = json.load(f)
        self.numPista = num_pista

        self.quantidade_controle = len(self.pistas[self.numPista]["Pontos_ctrl"])
        self.caminhoBaseModelos = caminho_modelos
        self.pontuacao = 0
        self.contagem_regressiva = 0
        self.pontuacao_total = 0
        self.pontuacao_maxima = 0
        self.terminado = 0
        self.mostrar_grafico = False
        self.qps = qps
        self.initModelo(modelo)
        self.agente.LimparMemoria()
        f.close()
        self.Reset()

    # Inicia o modelo com os parâmentros no arquivo config e carrega as informações do treinamento
    def initModelo(self, modelo):
        self.caminhoModelo = modelo
        self.lista_pontuacao = []
        self.pontuacao_guardar = []
        self.arq_pontuacao = f"{modelo}/Score"
        self.arq_pontuacao_maxima = f"{modelo}/HScore"
        self.caminho_imagem_grafico = f"{modelo}/imagens"
        arq_param = f"{modelo}/config"

        self.agente_config = utilidade.LerConfig(arq_param)
        if len(self.agente_config) == 0:
            print(f"ERRO {arq_param}")
            return

        self.eps_ini = float(self.agente_config[1])
        self.eps_decaimento = float(self.agente_config[2])
        self.eps_min = float(self.agente_config[3])

        self.agente = Agente(self.eps_ini, float(self.agente_config[4]), float(self.agente_config[5]), int(self.agente_config[6]), 
                           int(self.agente_config[7]), int(self.agente_config[8]), int(self.agente_config[9]), self.agente_config[10], modelo)

        if os.path.exists(modelo):
            pontuacoes, quantidade_pontuacao = utilidade.LerArquivo(self.arq_pontuacao)
            self.agente.n_partidas = len(pontuacoes)
            pontuacao_maxima,_ = utilidade.LerArquivo(self.arq_pontuacao_maxima)
            if self.agente.n_partidas > 0:
                self.AtualizaPontuacaoMaxima(pontuacao_maxima)
            self.agente.epsilon = max(self.eps_min, self.agente.epsilon*math.pow(self.eps_decaimento, self.terminado))
            self.lista_pontuacao.extend(pontuacoes)
            if os.path.exists(f"{modelo}/model.pth"):
                self.agente.modelo.Load(f"{modelo}/model.pth")
                self.agente.AtualizaAlvo()
        if not os.path.exists(modelo):
            os.makedirs(modelo)
        
        if self.agente.n_partidas % 100 == 0 and self.agente.n_partidas != 0:
            self.Reset(True)
        
        if os.path.exists(self.caminho_imagem_grafico):
            self.num_imagem = len(os.listdir(f"{self.caminho_imagem_grafico}/"))
        else:
            os.makedirs(self.caminho_imagem_grafico)
            self.num_imagem = 0


    def AtualizaPontuacaoMaxima(self, score):
        self.pontuacao_maxima = int(score[0])
        self.terminado = int(score[1])


    def Reset(self, teste=False):
        self.carro.reset(self.pistas[self.numPista]["Ponto_ini"][0][0], self.pistas[self.numPista]["Ponto_ini"][0][1])
        self.ponto_controle = self.pistas[self.numPista]["Ponto_ini"][0][2]
        self.dist_controle = None
        self.dist_controle_maxima = None
        self.pontuacao = 0
        self.contagem_regressiva = 0
    
    # Restaura a posição do carro
    def Reset_pos(self):
        self.carro.reset(self.pistas[self.numPista]["Ponto_ini"][0][0], self.pistas[self.numPista]["Ponto_ini"][0][1])
        self.dist_controle = None
        self.dist_controle_maxima = None

    # Verifica a colisão do carro com a pista
    def VerificaColisaoCarro(self, parede_ext, parede_int):
        if utilidade.ColisaoPoligono(self.carro.vector, parede_ext):
            return True

        if utilidade.ColisaoPoligono(self.carro.vector, parede_int):
            return True

    # Verifica a colisão das linhas de visão com a pista
    def VerificaColisaoLinha(self, parede_ext, parede_int):
        colisoes = []
        t1, t2 = [], []
        rt = []
        aux, t1 = utilidade.ColisaoPoligonoLinha(self.carro.linhas, parede_ext)
        rt.extend(t1)
        if aux != None:
            colisoes.extend(aux)
        aux, t2 = utilidade.ColisaoPoligonoLinha(self.carro.linhas, parede_int)
        if aux != None:
            for i in range(0, len(self.carro.linhas)):
                if t2[i] < t1[i]:
                   colisoes[i] = aux[i]
                   rt[i] = t2[i]

        return colisoes, rt

    # Verifica a collisão do carro com o ponto de controle
    def ColisaoPontoControle(self, ponto_controle):
        if utilidade.ColisaoPoligono(self.carro.vector, ponto_controle):
            return True
        
    # Verifica a colisão das linhas de visão com o ponto de controle
    def DetectarPontoControle(self, ponto_controle):
        colisoes, t1 = utilidade.ColisaoPoligonoLinha(self.carro.linhas, ponto_controle)
        return colisoes, t1
    
    # Verifica a distância entre o carro e o ponto de controle
    def DistanciaPontoControle(self, pontos_controle):
        meio_ponto = ((pontos_controle[0][0] + pontos_controle[1][0])/2, 
                      (pontos_controle[0][1] + pontos_controle[1][1])/2)
        
        meio_carro = ((self.carro.vector[0][0] + self.carro.vector[1][0])/2, (self.carro.vector[0][1] + self.carro.vector[1][1])/2)
        meio_tras_carro = ((self.carro.vector[2][0] + self.carro.vector[3][0])/2, (self.carro.vector[2][1] + self.carro.vector[3][1])/2)
        self.linha = [meio_ponto, meio_carro]
        dist = math.hypot((meio_ponto[0] - meio_carro[0]), (meio_ponto[1] - meio_carro[1]))
        dist_tras = math.hypot((meio_ponto[0] - meio_tras_carro[0]), (meio_ponto[1] - meio_tras_carro[1]))
        return dist, dist_tras

    # Calcula a recompensa do agente baseado na distância do ponto de controle
    def RecompensaDistancia(self):
        dist = 999
        recompensa = 0
        
        dist, dist_tras = self.DistanciaPontoControle(self.pistas[self.numPista]["Pontos_ctrl"][self.ponto_controle])
        if dist == 999:
            return 0
        if self.dist_controle == None or self.dist_controle == 999:
            self.dist_controle = dist
            self.dir_carro = 1
            return 0 
        if dist > dist_tras:
            recompensa -= 5
        if self.dist_controle == dist:
            self.dir_carro = 1
            return 0 + recompensa
        elif dist > self.dist_controle + 15:
            self.dir_carro = -1
            if self.dist_controle_maxima == None or self.dist_controle_maxima == 999 or self.dist_controle_maxima < dist:
                self.dist_controle_maxima = dist
                bonus = self.dist_controle - dist - 10
                return bonus + recompensa
            return 0 + recompensa
        elif dist < self.dist_controle:
            self.dir_carro = 1
            bonus = self.dist_controle - dist + 5
            self.dist_controle = dist
            self.dist_controle_maxima = dist
            return bonus + recompensa
        else:
            return 0 + recompensa

    # Calcula a punição do agente baseado de sua distância à parede
    def PunicaoParede(self, parede_ext, parede_int):
        dist = 999
        linhas = self.carro.linhas
        collision, _ = self.VerificaColisaoLinha(parede_ext, parede_int)
        for i in range(0, len(linhas)):
            if collision[i] != None:
                aux = math.hypot((linhas[i].vector[0][0] - collision[i][0]), 
                                 (linhas[i].vector[0][1] - collision[i][1]))
                if aux < dist:
                    dist = aux
        if dist <= 15:
            return dist - 20
        return 0
        

    def Draw(self, tela):
        if self.mostrar_grafico:
            self.grafico.draw(self.grafico_sur, self.lista_pontuacao, self.quantidade_controle, self.grafico_modificado, self.grafico_tipo)
            tela.blit(self.grafico_sur, [0,0])
            self.grafico_modificado = False
            self.botoes['stp_vis'].draw(tela)
            self.botoes['voltar'].draw(tela)
            self.botoes['print'].draw(tela)
            self.botoes['mudar_grafico'].draw(tela)
        else:
            tela.fill((0, 200, 0))
            pygame.draw.polygon(tela, (175, 175, 175), self.pistas[self.numPista]["Parede_ext"])
            pygame.draw.polygon(tela, (0, 200, 0), self.pistas[self.numPista]["Parede_int"])
            
            pygame.draw.polygon(tela, (255, 255, 255), self.pistas[self.numPista]["Parede_ext"], width=5)
            pygame.draw.polygon(tela, (255, 255, 255), self.pistas[self.numPista]["Parede_int"], width=5)

            linha = self.pistas[self.numPista]["Pontos_ctrl"][self.ponto_controle]
            pygame.draw.line(tela, (0, 0, 0), linha[0], linha[1], width=4)

            pygame.draw.line(tela, (255, 0, 0), self.linha[0], self.linha[1], width=2)

            for col in self.pontos_col:
                if col != None:
                    pygame.draw.circle(tela, (100, 255, 100), col, 5, 0)

            for col in self.col_check:
                if col != None:
                    pygame.draw.circle(tela, (255, 0, 255), col, 5, 0)

            self.carro.draw(tela)
            if self.mostrar_dados:
                texto = FONTE.render(f"Pontuação: {str(self.pontuacao_total)}/{str(self.pontuacao_maxima)}", True, (255, 255, 255))
                tela.blit(texto, [0, 0])
                texto = FONTE.render(f"Voltas: {self.terminado}", True, (255, 255, 255))
                tela.blit(texto, [0, 25])
                texto = FONTE.render(f"Partidas: {self.agente.n_partidas}", True, (255, 255, 255))
                tela.blit(texto, [0, 50])
                texto = FONTE.render(f"Epsilon: {self.agente.epsilon:.2}", True, (255, 255, 255))
                tela.blit(texto, [300, 0])
                self.botoes['esconde'].draw(tela)
            else:
                self.botoes['mostra'].draw(tela)

            if self.travar_qps:
                self.botoes['destrava_qps'].draw(tela)
            else:
                self.botoes['trava_qps'].draw(tela)
                
            self.botoes['ini_vis'].draw(tela)
            self.botoes['voltar'].draw(tela)

        pygame.display.update()

    # Verifica o estado atual do simulador
    def EstadoSimulador(self):
        estado = []
        self.pontos_col, linha_parede = self.VerificaColisaoLinha(self.pistas[self.numPista]["Parede_ext"], 
                                                                   self.pistas[self.numPista]["Parede_int"])
        self.col_check, linha_controle = self.DetectarPontoControle(self.pistas[self.numPista]["Pontos_ctrl"][self.ponto_controle])
        for i in range(len(linha_parede)):
            if linha_parede[i] != 999:
                linha_parede[i] = round(linha_parede[i]*100)
        for i in range(len(linha_controle)):
            if linha_controle[i] != 999:
                linha_controle[i] = round(linha_controle[i]*100)
        linhas = self.carro.linhas
        distancias = [-1] * (len(linhas)*2)
        for i in range(0, len(linhas)):
            if self.pontos_col[i] == None:
                distancias[i] = -1
            else:
                distancias[i] = round(math.hypot((linhas[i].vector[0][0] - self.pontos_col[i][0]), 
                                                 (linhas[i].vector[0][1] - self.pontos_col[i][1])), 2)
                

        deslocamento = i
        for i in range(0, len(linhas)):
            if self.pontos_col[i] == None:
                distancias[i+deslocamento] = -1
            else:
                distancias[i+deslocamento] = round(math.hypot((linhas[i].vector[0][0] - self.pontos_col[i][0]), 
                                                   (linhas[i].vector[0][1] - self.pontos_col[i][1])), 2)
        estado.extend(distancias)
        estado.append(self.carro.vel)
        estado.append(self.dir_carro)
        return estado

    # Executa um passo no simulador
    def PassoSimulador(self, tela):
        if self.travar_qps:
            self.clock.tick(self.qps)
        for evento in pygame.event.get():
            if evento.type == pygame.QUIT:
                if self.agente.n_partidas % 10 != 0:
                    self.SalvaModelo()
                return -2

        entrada = pygame.key.get_pressed()

        if entrada[pygame.K_ESCAPE]:
            if self.agente.n_partidas % 10 != 0:
                self.SalvaModelo()
            return -2

        if self.mostrar_grafico:
            if self.botoes['stp_vis'].Verificar(self.apertado):
                self.mostrar_grafico = False
            elif self.botoes['print'].Verificar(self.apertado):
                pygame.image.save(self.grafico_sur, f"{self.caminho_imagem_grafico}/imagem{self.num_imagem}.png")
            elif self.botoes['mudar_grafico'].Verificar(self.apertado):
                if self.grafico_tipo == 0:
                    self.grafico_tipo = 1
                elif self.grafico_tipo == 1:
                    self.grafico_tipo = 0
        else:
            if self.botoes['ini_vis'].Verificar(self.apertado):
                self.mostrar_grafico = True
                self.grafico_modificado = True
            elif self.botoes['trava_qps'].Verificar(self.apertado) and not self.travar_qps:
                self.travar_qps = True
            elif self.botoes['destrava_qps'].Verificar(self.apertado) and self.travar_qps:
                self.travar_qps = False
            elif self.botoes['mostra'].Verificar(self.apertado) and not self.mostrar_dados:
                self.mostrar_dados = True
            elif self.botoes['esconde'].Verificar(self.apertado) and self.mostrar_dados:
                self.mostrar_dados = False
        
        if self.botoes['voltar'].Verificar(self.apertado):
            if self.agente.n_partidas % 10 != 0:
                self.SalvaModelo()
            return -1
        

        if pygame.mouse.get_pressed()[0] and not self.apertado:
            self.apertado = True
        if not pygame.mouse.get_pressed()[0] and self.apertado:
            self.apertado = False
            
        estado_inicial = self.EstadoSimulador()
        acao = self.agente.GerarAcao(estado_inicial)
        self.Draw(tela)
        self.carro.Entrada(acao)
        recompensa = 0
        completo = False
        if self.VerificaColisaoCarro(self.pistas[self.numPista]["Parede_ext"], self.pistas[self.numPista]["Parede_int"]):
            completo = True
            recompensa = -30
            if self.pontuacao_total > self.pontuacao_maxima:
                self.pontuacao_maxima = self.pontuacao_total
        elif self.contagem_regressiva >= 600:
            completo = True
            recompensa = 0
            if self.pontuacao_total > self.pontuacao_maxima:
                self.pontuacao_maxima = self.pontuacao_total
        elif self.ColisaoPontoControle(self.pistas[self.numPista]["Pontos_ctrl"][self.ponto_controle]):
            self.pontuacao += 1
            self.pontuacao_total += 1
            recompensa = 50
            dist, dist_tras = self.DistanciaPontoControle(self.pistas[self.numPista]["Pontos_ctrl"][self.ponto_controle])
            if dist_tras < dist:
                recompensa += 0 if dist >= 20 else math.ceil((20 - dist)/2)
            else:
                recompensa -= 10
            recompensa += self.PunicaoParede(self.pistas[self.numPista]["Parede_ext"], self.pistas[self.numPista]["Parede_int"])
            self.contagem_regressiva = 0
            self.ponto_controle = (self.ponto_controle + 1)
            if self.ponto_controle >= self.quantidade_controle:
                self.ponto_controle = 0
                if self.pistas[self.numPista]["Reset_pos"]:
                    self.Reset_pos()
                self.terminado += 1
                self.agente.epsilon = max(self.eps_min, self.agente.epsilon*self.eps_decaimento)
                if self.pontuacao_total > self.pontuacao_maxima:
                    self.pontuacao_maxima = self.pontuacao_total
            self.dist_controle = None
            self.dist_controle_maxima = None
        else:
            recompensa += self.RecompensaDistancia()
            recompensa += self.PunicaoParede(self.pistas[self.numPista]["Parede_ext"], self.pistas[self.numPista]["Parede_int"])
            self.contagem_regressiva += 1

        novo_estado = self.EstadoSimulador()
        self.agente.Guardar (estado_inicial, acao, recompensa, novo_estado, completo)
        self.agente.TreinarMemoriaCurta()
        
        if completo:
            self.agente.n_partidas += 1
            if self.agente.n_partidas % 100 == 0:
                self.agente.AtualizaAlvo()
            perda = self.agente.TreinarMemoriaLonga()
            self.pontuacao_guardar.append(self.pontuacao_total)
            self.lista_pontuacao.append(self.pontuacao_total)
            if self.agente.n_partidas % 10 == 0:
                self.SalvaModelo()
            self.Reset()
            self.pontuacao_total = 0
            self.grafico_modificado = True
        return 0

    # Salva os resultados do treinamento e o modelo
    def SalvaModelo(self):
        utilidade.EscreverArquivo(self.arq_pontuacao, self.pontuacao_guardar)
        self.pontuacao_guardar.clear()
        utilidade.DelEscreverArquivo(self.arq_pontuacao_maxima, (self.pontuacao_maxima, self.terminado))
        self.agente.modelo.Save()
        if self.agente.n_partidas % 1000 == 0:
            if not os.path.exists(f"{self.caminhoModelo}/Backups/Geral"):
                os.makedirs(f"{self.caminhoModelo}/Backups/Geral")
                os.makedirs(f"{self.caminhoModelo}/Backups/Models")
            shutil.copy(self.arq_pontuacao, f"{self.caminhoModelo}/Backups/Geral/Score{int(abs(self.agente.n_partidas/1000))}")
            shutil.copy(self.arq_pontuacao_maxima, f"{self.caminhoModelo}/Backups/Geral/HScore{int(abs(self.agente.n_partidas/1000))}")
            shutil.copy(f"{self.caminhoModelo}/model.pth", 
                        f"{self.caminhoModelo}/Backups/Models/model{int(abs(self.agente.n_partidas/1000))}.pth")


    def Close(self):
        self.agente.modelo.Save()

