import pygame
import math
import os

"""Verifica se os pontos A, B e C estão em sentido anti-horário."""
def CCW(A,B,C):
    return (C[1]-A[1]) * (B[0]-A[0]) > (B[1]-A[1]) * (C[0]-A[0])

"""
    Dado duas linhas (A,B) e (C,D) se cruzam caso:
    - Os pontos (A,B,C) e (A,B,D) estão em sentidos diferentes;
    - Os pontos (C,D,A) e (C,D,B) estão em sentidos diferentes.
"""
def Cruzamento(A,B,C,D):
    return CCW(A,C,D) != CCW(B,C,D) and CCW(A,B,C) != CCW(A,B,D)

"""
    Verificando se dois poligonos colidem verificando se existe um cruzamento 
    entre as linhas.
"""
def ColisaoPoligono(poli1, poli2):
    ant1 = len(poli1)-1
    for ponto1 in range(0, len(poli1)):
        ant2 = len(poli2)-1
        for ponto2 in range(0, len(poli2)):
            if Cruzamento(poli1[ant1], poli1[ponto1], poli2[ant2], poli2[ponto2]):
                return True
            ant2 = ponto2
        ant1 = ponto1


"""
    Calcula a posição em que as linhas se cruzam onde:
    - linha1 e linha2 são dois segmentos de linha;
    - t é a posição do cruzamento na linha 1, onde t = 0 se refere a um cruzamento 
      no ponto inicial do segmento e t = 1 se refere ao ponto final, caso t < 0 ou 
      t > 1 o ponto do cruzamento não está no segmento;
    - u é a posição do cruzamento na linha2.
    Retorna o ponto do cruzamento e t.
"""
def CruzamentoLinha(linha1, linha2):
    t_dividendo = (linha2[1][0]-linha2[0][0])*(linha1[0][1]-linha2[0][1])-(linha2[1][1]-linha2[0][1])*(linha1[0][0]-linha2[0][0])
    u_dividendo = (linha2[0][1]-linha1[0][1])*(linha1[0][0]-linha1[1][0])-(linha2[0][0]-linha1[0][0])*(linha1[0][1]-linha1[1][1])
    divisor = (linha2[1][1]-linha2[0][1])*(linha1[1][0]-linha1[0][0])-(linha2[1][0]-linha2[0][0])*(linha1[1][1]-linha1[0][1])

    if divisor == 0:
        return None, None
    
    t = t_dividendo/divisor
    u = u_dividendo/divisor
    
    if t >= 0 and t <= 1 and u >= 0 and u <= 1:
        return [
            linha1[0][0]+(linha1[1][0]-linha1[0][0])*t,
            linha1[0][1]+(linha1[1][1]-linha1[0][1])*t
        ], t
    return None, None

"""
    Verificando se um poligono e um conjunto de linhas colidem verificando se 
    existe um cruzamento entre as linhas.
"""
def ColisaoPoligonoLinha(linhas, poli):
    colide = [None] * len(linhas)
    t = [999] * len(linhas)
    for linha in range(0, len(linhas)):
        ant = len(poli)-1
        for ponto in range(0, len(poli)):
            aux, taux = CruzamentoLinha(linhas[linha].vector, [poli[ant], poli[ponto]])
            if aux != None:
                if taux < t[linha]:
                    colide[linha] = aux
                    t[linha] = taux
            ant = ponto
    return colide, t

"""
    A classe usada para as linhas de visão do carro.
"""
class Linha:
    def __init__(self, pos, tamanho, angulo_original, angulo_carro):
        self.pos = []
        self.vector = [[pos[0], pos[1]], [-1, -1]]
        self.angulo_original = angulo_original
        self.tamanho = tamanho
        self.angulo = angulo_original + angulo_carro
        self.Calcular()
    
    #modifica o ângulo da linha
    def SomaAngulo(self, angulo):
        self.angulo = (self.angulo + angulo)%360
        self.Calcular()

    #modifica a posição da linha
    def MovePonto(self, pos):
        self.vector[0][0] = pos[0]
        self.vector[0][1] = pos[1]
        self.Calcular()

    #calcula os pontos da linha baseado no ângulo, posição e tamanho da linha
    def Calcular(self):
        self.vector[1][0] = self.vector[0][0] + (self.tamanho * math.cos(math.radians(self.angulo)))
        self.vector[1][1] = self.vector[0][1] - (self.tamanho * math.sin(math.radians(self.angulo)))

    def draw(self, win):
        pygame.draw.lines(win, (0, 255, 255), True, self.vector)
    
    #restaura a linha para uma posição e angulo expecífico
    def reset(self, pos, novo_angulo):
        self.angulo = self.angulo_original + novo_angulo
        self.vector[0][0] = pos[0]
        self.vector[0][1] = pos[1]
        self.Calcular()


############Arquivos######################

"""
    Lê o arquivo e retorna uma lista com todos os valores e a soma total dos números.
"""
def LerArquivo(caminho):
    arr = []
    total = 0
    if os.path.exists(caminho):
        arquivo = open(caminho, "r")
        texto = arquivo.read()
        if texto != None:
            for num in texto.split(","):
                if num == '':
                    continue
                if num == 'None':
                    arr.append(None)
                else:
                    arr.append(float(num))
                    total += float(num)
        arquivo.close()
    return arr, total

"""
    Lẽ o arquivo de configuração do modelo.
"""
def LerConfig(caminho):
    arr = []
    if os.path.exists(caminho):
        arquivo = open(caminho, "r")
        texto = arquivo.read()
        if texto != None:
            for param in texto.split(","):
                if param == '':
                    continue
                else:
                    arr.append(param)
        arquivo.close()
    return arr

"""
    Escreve a lista de numeros no final do arquivo.
"""
def EscreverArquivo(caminho, numeros):
    if os.path.exists(caminho):
        arquivo = open(caminho, "a")
    else:
        arquivo = open(caminho, "x")
    for numero in numeros:
        arquivo.write(f",{numero}")
    arquivo.close()
"""
    Deleta o conteúdo do arquivo e escreve o novo conteúdo
"""
def DelEscreverArquivo(caminho, numeros):
    if os.path.exists(caminho):
        arquivo = open(caminho, "w")
    else:
        arquivo = open(caminho, "x")
    arquivo.write(f"{numeros[0]},{numeros[1]}")
    arquivo.close()